# A simple numpy ring buffer for data + timestamps
# This was abstracted from the StreamWatcher class, as it is often useful on its own
import numpy as np

from dareplane_utils.logging.logger import get_logger

logger = get_logger(__name__)


class RingBuffer:
    """
    A simple numpy ring buffer for data and timestamps.

    This class implements a ring buffer using numpy arrays to store data and corresponding timestamps.
    It is designed to efficiently handle a fixed-size buffer with FIFO (First In, First Out) behavior.
    The buffer can be used to store and retrieve data samples along with their timestamps.

    Attributes
    ----------
    buffer : np.ndarray
        The data buffer.
    buffer_t : np.ndarray
        The time buffer.
    last_t : float
        The latest timestamp.
    curr_i : int
        The index of the latest data point in the buffer.
    logger : logging.Logger
        The logger used for warnings and debug messages.

    Parameters
    ----------
    shape : tuple[int, int]
        The shape of the buffer, which needs to be at least 2D (n_samples, n_features).
        Arbitrary further dimensions can be added.
    dtype : type, optional
        A numpy data type for the buffer. Defaults to np.float32.

    Examples
    --------

    >>> rb = RingBuffer(shape=(10, 3), dtype=np.float32)
    >>> samples = [np.random.rand(3) for _ in range(5)]
    >>> times = list(range(5))
    >>> rb.add_samples(samples, times)
    >>>
    >>> print(rb.buffer)
    [[0.6456422  0.6063579  0.46437985]
    [0.77531135 0.12675048 0.10201751]
    [0.05525873 0.5852236  0.09854987]
    [0.7515863  0.20880483 0.7236128 ]
    [0.7157382  0.55806357 0.6247236 ]
    [0.         0.         0.        ]
    [0.         0.         0.        ]
    [0.         0.         0.        ]
    [0.         0.         0.        ]
    [0.         0.         0.        ]]
    >>>
    >>> print(rb.buffer_t)
    [0. 1. 2. 3. 4. 0. 0. 0. 0. 0.]
    >>>
    >>> print(rb.unfold_buffer())  # most recent sample is index=-1, second to most recent index=-2, etc.
    [[0.         0.         0.        ]
    [0.         0.         0.        ]
    [0.         0.         0.        ]
    [0.         0.         0.        ]
    [0.         0.         0.        ]
    [0.6456422  0.6063579  0.46437985]
    [0.77531135 0.12675048 0.10201751]
    [0.05525873 0.5852236  0.09854987]
    [0.7515863  0.20880483 0.7236128 ]
    [0.7157382  0.55806357 0.6247236 ]]
    >>>
    >>> print(rb.unfold_buffer_t())
    [0. 0. 0. 0. 0. 0. 1. 2. 3. 4.]


    >>> samples2 = [np.random.rand(3) for _ in range(8)]
    >>> times2 = list(range(8))
    >>> rb.add_samples(samples2, times2)
    >>> print(rb.buffer_t)
    [5. 6. 7. 3. 4. 0. 1. 2. 3. 4.]
    >>>
    >>> print(rb.unfold_buffer_t())
    [3. 4. 0. 1. 2. 3. 4. 5. 6. 7.]

    """

    def __init__(self, shape: tuple[int, int], dtype: type = np.float32):
        """
        Initialize the RingBuffer with a given shape and data type.

        Parameters
        ----------
        shape : tuple[int, int]
            The shape of the buffer, which needs to be at least 2D (n_samples, n_features).
            Arbitrary further dimensions can be added.
        dtype : type, optional
            A numpy data type for the buffer. Defaults to np.float32.
        """

        # Using numpy buffers
        self.buffer = np.zeros(shape, dtype=dtype)
        self.buffer_t = np.zeros(shape[0])
        self.last_t = 0  # last time stamp
        self.curr_i = 0
        self.logger = logger

    def legacy_add_samples(self, samples: list, times: list):
        if len(samples) > 0 and len(times) > 0:
            buffer_size = self.buffer.shape[0]
            if len(samples) > buffer_size:
                self.logger.warning(
                    "Received more data than fitting into the"
                    " buffer. Will only add data to fill buffer"
                    " once with the latest data"
                )
                samples = samples[-buffer_size:]
                times = times[-buffer_size:]
            # make it a ring buffer with FIFO
            old_i = self.curr_i
            self.curr_i = (self.curr_i + len(samples)) % buffer_size
            # self.logger.debug(f"{old_i=}, {self.curr_i=}, {len(samples)=}")
            # plain forward fill
            if old_i < self.curr_i:
                self.buffer[old_i : self.curr_i] = samples
                self.buffer_t[old_i : self.curr_i] = times
            # fill buffer up
            elif self.curr_i == 0:
                self.buffer[old_i:] = samples
                self.buffer_t[old_i:] = times
            # split needed -> start over at beginning
            else:
                self.logger.debug("Splitting data to add as buffer is full")
                nfull = buffer_size - old_i
                self.buffer[old_i:] = samples[:nfull]
                self.buffer_t[old_i:] = times[:nfull]
                self.buffer[: self.curr_i] = samples[nfull:]
                self.buffer_t[: self.curr_i] = times[nfull:]

            self.last_t = times[-1]

    def unfold_buffer(self):
        """
        Unfold the buffer to return the data in chronological order.

        This method returns the data in the buffer in chronological order, ending with the most recent sample.

        Returns
        -------
        np.ndarray
            The unfolded data buffer.
        """
        return np.vstack([self.buffer[self.curr_i :], self.buffer[: self.curr_i]])

    def unfold_buffer_t(self):
        """
        Unfold the buffer_t (time stamp buffer) to return time stamps in chronological order.

        This method returns the time stamps in the buffer in chronological order, ending with the most recent time stamp.

        Returns
        -------
        np.ndarray
            The unfolded data buffer.
        """
        # Do hstack here as the time buffer will be of dim (n,1) anyways
        return np.hstack([self.buffer_t[self.curr_i :], self.buffer_t[: self.curr_i]])

    def get_insert_slices(
        self, len_samples: int
    ) -> tuple[list[slice], list[slice], int]:
        """Get slices mapping data from the samples to the buffer

        Parameters
        ----------
        len_samples : int
            number of samples to add

        Returns
        -------
        tuple[list[slice], list[slice], int]

        """
        old_i = self.curr_i
        new_i = (self.curr_i + len_samples) % self.buffer.shape[0]

        if old_i < new_i:
            return [slice(old_i, new_i)], [slice(0, len_samples)], new_i
        elif new_i == 0:
            return [slice(old_i, None)], [slice(0, len_samples)], new_i
        else:
            nfull = self.buffer.shape[0] - old_i
            return (
                [slice(old_i, None), slice(0, new_i)],
                [slice(0, nfull), slice(nfull, None)],
                new_i,
            )

    def add_samples(self, samples: list, times: list):
        """Add samples to the buffer and progress the index"""

        if len(samples) == 0 or len(times) == 0:
            self.logger.warning(
                f"Received empty data {samples=}, {times=}, " "not adding to buffer"
            )
        else:
            buffer_size = self.buffer.shape[0]
            if len(samples) > buffer_size:
                self.logger.warning(
                    "Received more data than fitting into the"
                    " buffer. Will only add data to fill buffer"
                    " once with the latest data"
                )
                samples = samples[-buffer_size:]
                times = times[-buffer_size:]

            # make it a ring buffer with FIFO
            # This step should take about 300ns for 1024 samples
            slice_buffer, slice_samples, self.curr_i = self.get_insert_slices(
                len(samples)
            )

            if len(slice_buffer) > 1:
                self.add_split_buffer(slice_buffer, slice_samples, samples, times)

            else:
                self.add_continuous_buffer(slice_buffer, samples, times)

            self.last_t = times[-1]

    # create a lot of smaller function calls for profiling
    def add_split_buffer(self, slice_buffer, slice_samples, samples, times):
        # self.logger.debug("Splitting data to add as buffer is full")
        self.buffer[slice_buffer[0]] = samples[slice_samples[0]]
        self.buffer[slice_buffer[1]] = samples[slice_samples[1]]
        self.buffer_t[slice_buffer[0]] = times[slice_samples[0]]
        self.buffer_t[slice_buffer[1]] = times[slice_samples[1]]

    def add_continuous_buffer(self, slice_buffer, samples, times):
        """
        Slice samples should not be necessary >>> as we add continuously
        + slice selection from lists is slow
        """
        self.buffer[slice_buffer[0]] = samples
        self.buffer_t[slice_buffer[0]] = times


if __name__ == "__main__":
    rb = RingBuffer(shape=(10, 3), dtype=np.float32)
    samples = [np.random.rand(3) for _ in range(5)]
    times = list(range(5))

    rb.add_samples(samples, times)

    print(rb.buffer)
    print(rb.buffer_t)
    print(
        rb.unfold_buffer()
    )  # most recent sample is index=-1, second to most recent index=-2, etc.
    print(rb.unfold_buffer_t())

    samples2 = [np.random.rand(3) for _ in range(8)]
    times2 = list(range(8))
    rb.add_samples(samples2, times2)

    print(rb.buffer_t)
    print(rb.unfold_buffer_t())



import logging
import select
import socketserver
import struct
from pathlib import Path

import ujson
from fire import Fire

from dareplane_utils.logging.logger import default_dareplane_config

# NOTE: For a usual Dareplane setup, we will not have more than one
# instance of a LogRecordSocketReceiver, which is usually managed via the
# control room module. It is still included here next to the logger and the
# ujson_socket_handler, as it is required for unit testing and would allow
# an easier testing of multiple modules (still logging to a single file)
# if the control room would not be used.


# Basically taken from here: https://docs.python.org/2/howto/logging-cookbook.html#sending-and-receiving-logging-events-across-a-network
class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            try:
                chunk = self.connection.recv(4)
            except ConnectionResetError:
                # Remote host closed the connection
                break

            if len(chunk) < 4:
                break

            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))

            obj = self.load_ujson(chunk)

            record = logging.makeLogRecord(obj)

            # Overwrite as args sent from plotly dash errors would otherwise
            # break at logger.handle(record)
            if record.args == "()":
                record.args = []

            # Other adjustments for records emitted via dash/waitress
            if isinstance(record.created, str):
                record.created = float(record.created)

            if isinstance(record.relativeCreated, str):
                record.relativeCreated = float(record.relativeCreated)

            if isinstance(record.msecs, str):
                record.msecs = float(record.msecs)

            if "%s" in record.msg:
                record.msg = record.msg % record.args  # replacing %s with args for dash

            self.handle_log_record(record)

            # Example record from dash / waitress:
            # {
            #     "name": "control_room.gui.app",
            #     "msg": "Exception on /_dash-update-component [POST]",
            #     "args": "()",
            #     "levelname": "ERROR",
            #     "levelno": 40,
            #     "pathname": "/Users/../wvenv3.12/lib/python3.12/site-packages/flask/app.py",
            #     "filename": "app.py",
            #     "module": "app",
            #     "exc_info": "(<class 'KeyError'>, KeyError('data_root'), <traceback object at 0x106eab100>)",
            #     "exc_text": "None",
            #     "stack_info": "None",
            #     "lineno": 828,
            #     "funcName": "log_exception",
            #     "created": "1713889360.199748",
            #     "msecs": "199.0",
            #     "relativeCreated": "20730.786085128784",
            #     "thread": 6158839808,
            #     "threadName": "waitress-0",
            #     "processName": "MainProcess",
            #     "process": 57978,
            #     "taskName": "None",
            # }

    def load_ujson(self, data):
        return ujson.loads(data)

    def handle_log_record(self, record: logging.LogRecord):
        # if a name is specified, we use the named logger rather than the one
        # implied by the record.
        if self.server.logname is not None:
            name = self.server.logname
        else:
            name = record.name

        logger = logging.getLogger(name)

        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    Simple TCP socket-based logging receiver suitable for testing.
    """

    allow_reuse_address = 1

    def __init__(
        self,
        host="localhost",
        port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,  # 9020
        handler=LogRecordStreamHandler,
        logfile: Path = Path("default_socket.log"),
    ):
        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.abort = 0
        self.timeout = 1
        self.logname = None
        self.logfile = logfile

        # All derived loggers within this process will log to a file
        modify_root_logger(self.logfile)

    def serve_until_stopped(self):
        print("LogRecordSocketReceiver is up and listening")
        abort = 0

        while not abort:
            rd, wr, ex = select.select([self.socket.fileno()], [], [], self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort


def modify_root_logger(logfile: Path):
    cfg = default_dareplane_config.copy()
    # Set the target log file
    cfg["handlers"]["file"] = {
        "class": "logging.FileHandler",
        "filename": str(logfile.resolve()),
        "formatter": "dareplane_standard",
    }

    # Filtering should be done at client level - server will log all
    cfg["root"]["level"] = logging.DEBUG
    cfg["root"]["handlers"] = ["console", "file"]

    logging.config.dictConfig(cfg)


def main(logfile: str = "default_socket.log"):
    rcv = LogRecordSocketReceiver(logfile=Path(logfile))
    rcv.serve_until_stopped()


if __name__ == "__main__":
    Fire(main)



import socket
import subprocess
import threading
import time
from dataclasses import dataclass, field
from logging import Logger
from typing import Callable

from dareplane_utils.default_server.functions import (
    interpret_msg,
    stop_process,
    stop_thread,
)
from dareplane_utils.general.time import sleep_s
from dareplane_utils.logging.logger import get_logger


class UnknownMsgInterpretation(Exception):
    """Raised when the interpretation of a message is unknown"""

    pass


@dataclass
class DefaultServer:
    """
    A class representing a default server used and modified by other Dareplane projects.
    This server handles incoming TCP connections, interprets messages, and manages threads and processes.

    Attributes
    ----------
    port : int
        The port number on which the server listens. Default is 8080.
    ip : str
        The IP address on which the server listens. Default is "0.0.0.0".
    nlisten : int
        The maximum number of queued connections. Default is 10.
    name : str
        The name of the server. Default is "default_server".
    thread_stopper : Callable
        A function to stop threads. Default is `stop_thread`.
    proc_stopper : Callable
        A function to stop processes. Default is `stop_process`.
    msg_interpreter : Callable
        A function to interpret messages. Default is `interpret_msg`.
    pcommand_map : dict
        A dictionary mapping commands to their handlers.
    current_conn : socket.socket | None
        The current active connection.
    server_socket : socket.socket | None
        The server socket.
    threads : dict[str, tuple[threading.Thread, threading.Event]]
        A dictionary of active threads and their stop events.
    processes : dict[str, subprocess.Popen]
        A dictionary of active subprocesses.
    is_listening : bool
        A flag indicating whether the server is currently listening.
    logger : Logger
        The logger for the server.
    """

    port: int = 8080
    ip: str = "0.0.0.0"
    nlisten: int = 10
    name: str = "default_server"
    thread_stopper: Callable = stop_thread
    proc_stopper: Callable = stop_process
    msg_interpreter: Callable = interpret_msg
    pcommand_map: dict = field(default_factory=dict)
    current_conn: socket.socket | None = None
    server_socket: socket.socket | None = None
    threads: dict[str, tuple[threading.Thread, threading.Event]] = field(
        default_factory=dict
    )
    processes: dict[str, subprocess.Popen] = field(default_factory=dict)
    is_listening: bool = False
    logger: Logger = get_logger(__name__)

    def init_server(self, stop_event: threading.Event = threading.Event()):
        """
        Initialize the server socket and set up the stop event.

        This method creates a socket, sets socket options, binds the socket to the specified IP and port,
        and sets it to listen for incoming connections. It also initializes a threading event to control
        the server's listening state.

        Parameters
        ----------
        stop_event : threading.Event, optional
            A threading event to control the server's listening state. Default is a new threading.Event.

        """
        # spawn a socket
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.settimeout(None)
        server_socket.bind((self.ip, self.port))
        server_socket.listen(self.nlisten)
        self.server_socket = server_socket

        # adding a threading style stop event to potentially run listening in a separate thread
        # and still being able to stop from the controlling script. Otherwise only the socket client could send a stop
        self.listen_stop_event: threading.Event = stop_event

    def start_listening(self):
        """
        Start the server and listen for incoming connections.

        This method enters a loop where it waits for incoming TCP connections. When a connection is accepted,
        it handles incoming messages until the server is stopped. The method also manages the server's listening state
        and logs relevant information.
        """

        # main loop for running the server
        while not self.listen_stop_event.is_set():
            self.is_listening = True
            # for now just work on one / the current connection, implement
            # TODO: implement dealing with multiple connection
            try:
                current_conn, addr = self.server_socket.accept()  # type: ignore
            except Exception as err:
                self.logger.error(
                    f"Error accepting connection at {self.ip=}, {self.port=}, {self.server_socket=}"
                )
                raise err

            self.current_conn = current_conn
            self.current_conn.sendall(f"Connected to {self.name}\n".encode())
            while not self.listen_stop_event.is_set():
                try:
                    msg = self.current_conn.recv(2048)
                    if msg:
                        self.handle_msg(msg)

                except socket.timeout as err:
                    self.logger.info(f"Caugth timeout error {err=}")

                except KeyboardInterrupt as _:
                    self.logger.info("Received KeyboardInterrupt - stopping the server")

                except UnicodeDecodeError as _:
                    self.current_conn.sendall(
                        f"Was unable to decode {msg=} to ascii\n".encode()  # type: ignore
                    )
                except ConnectionResetError as err:
                    self.logger.info("Connection was reset by host- stopping the server")
                    self.is_listening = False
                    raise err
                except Exception as err:
                    self.logger.error(f"Caught error {err=}")
                    self.is_listening = False
                    raise err

        self.is_listening = False

    def handle_msg(self, msg: bytes):
        """
        Interpret the incoming message.

        This method processes the incoming message, checks if it is a concatenation of multiple commands,
        and handles each command individually. It also logs the received message and handles any errors that occur
        during message processing.

        Parameters
        ----------
        msg : bytes
            The incoming message to be interpreted.
        """

        self.logger.info(f"Received: {msg}")
        # Check if msg is concatenation of multiple commands, happends if
        # processing of a single command takes to long so multiple commands
        # end up at the socket
        if b";" in msg:
            for pc in msg.split(b";"):
                if pc != b"":
                    self.handle_msg(pc)
        else:

            # Default functionality which should always be there
            # and the same for all servers
            is_default_command = self.default_msg_interpretation(msg)

            # if not a default command, check the pcmommand map
            if not is_default_command:
                self.logger.debug("Interpreting non-default message")

                msg = msg.replace(b"\r\n", b"")
                # common start byte, would lead to an error in decode otherwise # noqa
                msg = msg.replace(b"\xc2", b"")
                # Ignore blanks -> e.g. accidental return in telnet
                if (
                    msg != b""
                    and msg.decode("ascii").split("|")[0] in self.pcommand_map.keys()
                ):
                    self.msg_interpretation(msg)  # type: ignore

                else:
                    self.logger.warning(f"Unknown pcomm in {msg=}")

    def default_msg_interpretation(self, msg: bytes) -> bool:
        """
        Interpret default messages.

        This method handles default commands such as stopping processes or threads, closing the server,
        and retrieving the list of available commands. It logs relevant information and performs the necessary actions
        based on the received message.

        Parameters
        ----------
        msg : bytes
            The incoming message to be interpreted.

        Returns
        -------
        bool
            True if the message was a default command, False otherwise.
        """
        if (
            msg == b"STOP\r\n"
            or msg == b"STOP"
            or msg == b"STOP|\r\n"
            or msg == b"STOP|"
        ):
            # Stop any processes or threads which could have been spawend by interpreting messages
            self.close_threads()
            self.close_processes()

        # Closing the server
        elif (
            msg == b"CLOSE\r\n"
            or msg == b"CLOSE"
            or msg == b"CLOSE|\r\n"
            or msg == b"CLOSE|"
        ):
            # stop listening for TCP connections and traffic
            self.listen_stop_event.set()

        elif b"GET_PCOMMS" in msg:
            msg = msg.replace(b"\r\n", b"")
            msg = msg.replace(
                b"\xc2", b""
            )  # common start byte, would lead to an error in decode otherwise

            self.current_conn.sendall(  # type: ignore
                ("|".join(list(self.pcommand_map.keys()) + ["STOP", "CLOSE"])).encode()
            )
        elif msg == b"UP":
            self.current_conn.sendall(b"1")  # type: ignore
        else:
            # Nothing was done
            return False

        return True

    def msg_interpretation(self, msg: str):
        """Interpret the message and perform book keeping if necessary"""
        ret = interpret_msg(msg, self.pcommand_map, logger=self.logger)

        # the book keeping part
        if isinstance(ret, int):
            self.logger.debug(f"{msg=} returned {ret=} after interpretation")
            pass

        # any implementation returning a thread should return the stop_event alongside
        # TODO: think about how this can be enforced
        elif (
            isinstance(ret, tuple)
            and isinstance(ret[0], threading.Thread)
            and isinstance(ret[1], threading.Event)
        ):
            self.logger.debug(f"{msg=} returned a thread after interpretation")
            self.threads[f"{time.time_ns()}_{msg}"] = ret

        elif isinstance(ret, subprocess.Popen):
            self.logger.debug(f"{msg=} returned a subprocess after interpretation")
            self.processes[f"{time.time_ns()}_{msg}"] = ret

        else:
            raise UnknownMsgInterpretation(
                "Unknown msg interpretation return for "
                f"{msg=} and {ret=} "
                " valid interpretations should return int|tuple[threading.Thread, threading.Event]|subprocess.Popen"
            )

    def shutdown(self):
        """Shutdown the server and close all connections"""
        self.logger.info(f"Shutting down {self.name}")
        if self.current_conn:
            self.current_conn.close()

        if self.server_socket:
            self.server_socket.close()
        self.close_threads()
        self.close_processes()

    def close_threads(self):
        # clean up potential threads
        self.logger.debug(f"Closing threads: {self.threads.items()}")
        removed_threads = []
        for thid, (th, stop_event) in self.threads.items():
            stop_event.set()
            self.thread_stopper(th, logger=self.logger)
            removed_threads.append(thid)
        for k in removed_threads:
            self.threads.pop(k, None)

    def close_processes(self):
        # clean up potential subprocesses
        self.logger.debug(f"Closing processes: {self.processes.items()}")
        removed_sprocesses = []
        for spid, sp in self.processes.items():
            self.proc_stopper(sp, logger=self.logger)
            removed_sprocesses.append(spid)
        for k in removed_sprocesses:
            self.processes.pop(k, None)

    def __del__(self):
        # also use the shutdown in the finalizer in case the instance is
        # removed from memory
        self.shutdown()


@dataclass
class DefaultCallbackServer(DefaultServer):
    """
    DefaultCallbackServer

    The default server extended with a callback capability.

    This class extends the DefaultServer class to include a callback mechanism. It allows the server to send
    callback messages to connected clients. The callback messages are stored in a stack and sent to the client
    in a separate thread.

    Attributes
    ----------
    callback_stack : list[str]
        A list of callback payloads, which will be sent to connected clients.
    """

    callback_stack: list[str] = field(
        default_factory=list
    )  # list of callback payloads, which will be sent to connected clients

    def start_listening(self):
        # main loop for running the server
        self.logger.debug("Callback server starting to listen")
        while not self.listen_stop_event.is_set():
            self.is_listening = True
            # for now just work on one / the current connection, implement
            # TODO: refactor this part for a cleaner implementation of the dealing with callbacks -> all copy and paste from above, besides 4 lines
            try:
                current_conn, addr = self.server_socket.accept()

                # Initialize the callback thread
                callback_stop_event = threading.Event()
                callback_thread = threading.Thread(
                    target=process_callbacks,
                    args=(self, callback_stop_event),
                )
                # add to automatically manage closing
                self.threads["callbacks"] = (
                    callback_thread,
                    callback_stop_event,
                )
                callback_thread.start()

            except Exception as err:
                self.logger.error(
                    f"Error accepting connection at {self.ip=}, {self.port=}, "
                    f"{self.server_socket=}"
                )
                raise err

            self.current_conn = current_conn
            self.current_conn.sendall(f"Connected to {self.name}\n".encode())
            while not self.listen_stop_event.is_set():
                try:
                    # Process incoming
                    msg = self.current_conn.recv(1024)
                    if msg:
                        self.logger.info(f"Received: {msg}")
                        # interpret the message

                        # Default functionality which should always be there
                        # and the same for all servers
                        is_default_command = self.default_msg_interpretation(msg)

                        # if not a default command, check the pcmommand map
                        if not is_default_command:
                            msg = msg.replace(b"\r\n", b"")
                            # common start byte, would lead to an error in decode otherwise # noqa
                            msg = msg.replace(b"\xc2", b"")
                            # Ignore blanks -> e.g. accidental return in telnet
                            if (
                                msg != b""
                                and msg.decode("ascii").split("|")[0]
                                in self.pcommand_map.keys()
                            ):
                                self.msg_interpretation(msg)

                            else:
                                self.logger.warning(f"Unknown pcomm in {msg=}")

                except socket.timeout as err:
                    self.logger.info(f"Caugth timeout error {err=}")

                except KeyboardInterrupt as _:
                    self.logger.info("Received KeyboardInterrupt - stopping the server")

                except UnicodeDecodeError as _:
                    self.current_conn.sendall(
                        f"Was unable to decode {msg=} to ascii\n".encode()  # type: ignore
                    )
                except Exception as err:
                    self.logger.error(f"Caught error {err=}")
                    self.is_listening = False
                    raise err

        self.is_listening = False


def process_callbacks(server: DefaultCallbackServer, stop_event: threading.Event):
    """
    Process callback messages and send them to the connected client.

    This function runs in a separate thread and continuously checks the callback stack for messages.
    If there are messages in the stack, it sends them to the connected client and removes them from the stack.
    The function stops processing when the stop event is set.

    Parameters
    ----------
    server : DefaultCallbackServer
        The server instance that contains the callback stack and the current connection.
    stop_event : threading.Event
        An event that, when set, signals the function to stop processing callbacks.
    """

    # Process callback stack
    while not stop_event.is_set():
        while len(server.callback_stack) > 0:
            server.current_conn.sendall(server.callback_stack.pop(0).encode())  # type: ignore
            sleep_s(0.001)



import numpy as np

from dareplane_utils.general.ringbuffer import RingBuffer


def test_ringbuffer_init():
    rb = RingBuffer(shape=(10, 3))
    assert rb.buffer.dtype == np.float32
    assert rb.buffer.shape == (10, 3)
    assert (rb.buffer == 0).all()


def test_adding_data_to_ringbuffer():
    rb = RingBuffer(shape=(10, 3))
    rb2 = RingBuffer(shape=(10, 3))

    rb.add_samples(np.ones((5, 3), dtype=np.float32), np.arange(5))

    assert np.all(rb.buffer[:5] == 1)
    assert np.all(rb.buffer[5:] == 0)

    # buffers should not be linked
    assert (rb2.buffer == 0).all()

    # check wrapping around
    rb.add_samples(np.ones((7, 3), dtype=np.float32) * 3, np.arange(7))
    assert np.all(rb.buffer[:2] == 3)
    assert np.all(rb.buffer[5:] == 3)
    assert np.all(rb.buffer[2:5] == 1)


def test_unfolding_ringbuffer():
    rb = RingBuffer(shape=(10, 3))

    rb.add_samples(np.ones((5, 3), dtype=np.float32), np.arange(5))

    d = rb.unfold_buffer()

    assert np.all(d[-5:] == rb.buffer[:5])


def test_multi_dimensional_buffer_adding():

    rb = RingBuffer(shape=(10, 3, 2))

    assert rb.buffer.shape == (10, 3, 2)

    x = np.ones((5, 3, 2), dtype=np.float32)
    rb.add_samples(x, np.arange(5))

    assert np.all(rb.buffer[:5] == 1)

    x = np.ones((7, 3, 2), dtype=np.float32) * 3
    rb.add_samples(x, np.arange(7))

    assert np.all(rb.buffer[:2] == 3)
    assert np.all(rb.buffer[5:] == 3)
    assert np.all(rb.buffer[2:5] == 1)


def test_multi_dimensional_buffer_unfold():

    rb = RingBuffer(shape=(10, 3, 2))

    x = np.ones((15, 3, 2), dtype=np.float32) * np.arange(15).reshape(-1, 1, 1)

    rb.add_samples(x, np.arange(15))

    assert rb.unfold_buffer().shape == (10, 3, 2)
    assert np.all(rb.unfold_buffer()[-1, :, :] == 14)



import socket
import threading
import time

import pytest

from dareplane_utils.default_server.server import DefaultServer
from dareplane_utils.logging.logger import get_logger
from tests.resources.shared import get_test_thread

logger = get_logger("testlogger")


# for checking which process is running at a given port, we can use netstat with e.g.
# `netstat -anv -p tcp | grep 8080`


@pytest.fixture
def get_default_server_with_thread_spawning():
    # Init the server
    server = DefaultServer(port=8080)
    stop_event = threading.Event()
    server.init_server(stop_event=stop_event)

    # Add process creation capabilities
    pcommand_map = {"STARTTHREAD": get_test_thread}
    server.pcommand_map = pcommand_map

    # Run the servers processing in a separate thread
    stop_event.clear()  # in case it was set before
    server_thread = threading.Thread(target=server.start_listening)
    server_thread.start()

    yield server_thread, stop_event, server

    # Teardown
    stop_event.set()
    server_thread.join()
    server.shutdown()


def test_spawning_thread_from_client(
    get_default_server_with_thread_spawning,
):
    server_thread, stop_event, server = get_default_server_with_thread_spawning

    # Send a message to the server to spawn a process
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect(("localhost", 8080))
    client.sendall(b"STARTTHREAD")

    time.sleep(0.1)
    logger.debug(f"{server.threads=}")

    # the process should be registerd for book keeping
    assert len(server.threads.keys()) == 1

    client.sendall(b"CLOSE")
    client.close()


def test_stopping_processes(get_default_server_with_thread_spawning):
    server_thread, stop_event, server = get_default_server_with_thread_spawning

    # Send a message to the server to spawn a process
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect(("localhost", 8080))
    client.sendall(b"STARTTHREAD")

    time.sleep(0.1)
    server.close_threads()

    assert len(server.threads.keys()) == 0
    client.sendall(b"CLOSE")
    client.close()



import logging
import logging.config
import warnings

default_dareplane_config = {
    "version": 1,
    "formatters": {
        "dareplane_standard_color": {
            "()": "colorlog.ColoredFormatter",
            "format": "%(log_color)s%(asctime)s - %(name)s - %(levelname)s - %(reset)s %(white)s%(message)s",
        },
        "dareplane_standard": {
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "dareplane_standard_color",
        },
        "socket": {
            "formatter": "dareplane_standard",
            "class": "dareplane_utils.logging.ujson_socket_handler.UJsonSocketHandler",
            "host": "127.0.0.1",
            "port": 9020,
        },
    },
    "root": {
        "handlers": ["console", "socket"],
    },
}

# overwriting defaults
colors = {"DEBUG": "cyan"}

_config_applied = False


# have this as a simple wrapper to ensure the updated config is used
def get_logger(
    name: str,
    add_console_handler: bool = False,
    colors: dict = colors,
    no_socket_handler: bool = False,  # opt out for socket handler / TCP streaming
) -> logging.Logger:
    """
    Get a configured logger.

    Parameters
    ----------
    name : str
        The name of the logger.
    add_console_handler : bool, optional
        If True, add a console handler to the logger (default is False).
    colors : dict, optional
        A dictionary of colors for log levels (default is `colors`).
    no_socket_handler : bool, optional
        If True, opt out of adding a socket handler for TCP streaming (default is False).

    Returns
    -------
    logging.Logger
        The configured logger.
    """
    global _config_applied
    if not _config_applied:
        logging.config.dictConfig(default_dareplane_config)
        _config_applied = True
    
    logger = logging.getLogger(name)
    root_logger = logging.getLogger()

    if add_console_handler:
        consol_handler = [
            h for h in root_logger.handlers if isinstance(h, logging.StreamHandler)
        ][0]
        consol_handler.formatter.log_colors.update(colors)
        logger.addHandler(consol_handler)  # add console handler

    socket_handlers = [
        h for h in root_logger.handlers if isinstance(h, logging.handlers.SocketHandler)
    ]
    if socket_handlers:
        socket_handler = socket_handlers[0]
        
        # Warn once if socket handler cannot connect (helps debugging)
        if not hasattr(socket_handler, '_connection_warned'):
            try:
                # Test if we can create a socket connection
                if socket_handler.sock is None:
                    test_sock = socket_handler.makeSocket()
                    if test_sock:
                        test_sock.close()
            except (ConnectionRefusedError, OSError, Exception):
                warnings.warn(
                    f"Logging server not available at {socket_handler.host}:{socket_handler.port}. "
                    "Network logs will be lost. Logs will only appear in console.",
                    RuntimeWarning,
                    stacklevel=2
                )
            socket_handler._connection_warned = True
        
        logger.addHandler(socket_handler)  # add socket handler

    if no_socket_handler:
        logger.handlers = [
            h
            for h in logger.handlers
            if not isinstance(h, logging.handlers.SocketHandler)
        ]

    # do not propergate messages to root logger
    logger.propagate = False

    return logger



import sys
import time
from threading import Event

import numpy as np
import pytest

from dareplane_utils.general.event_loop import EventLoop
from dareplane_utils.logging.logger import get_logger

logger = get_logger(
    "dareplane_utils_test", add_console_handler=True, no_socket_handler=True
)
logger.setLevel("DEBUG")

# Windows Python 3.10 has poor timer resolution (~15ms)
# Python 3.11+ significantly improved this
IS_WINDOWS_PY310 = sys.platform == "win32" and sys.version_info[:2] == (3, 10)


def test_adding_non_callable_for_callback():
    ev = EventLoop(dt_s=0.01)
    with pytest.raises(AssertionError):
        ev.add_callback(1)

    with pytest.raises(AssertionError):
        ev.add_callback([1])

    with pytest.raises(AssertionError):
        ev.add_callback({"a": 2})


def test_missing_ctx_kwarg_in_callback():
    ev = EventLoop(dt_s=0.01)

    def test_foo():
        pass

    # should not lead to an error anymore, as we automatically wrap for convenience
    ev.add_callback(test_foo)

    def test_foo_ok(ctx={}):
        pass

    ev.add_callback(test_foo)


@pytest.mark.parametrize("dt_s", [0.001, 0.01, 0.1])
def test_accuracy_of_event_loop_for_different_dt(dt_s):

    buffer = np.zeros(max(100, int(1 / dt_s)))
    stop_event = Event()

    def cb(ctx: dict = {}, stop_event: Event = stop_event):
        buffer[ctx["idx"]] = time.perf_counter()
        ctx["idx"] = ctx["idx"] + 1
        if ctx["idx"] == len(buffer):
            stop_event.set()

    ev = EventLoop(dt_s=dt_s, stop_event=stop_event, ctx={"idx": 0})
    ev.add_callback(cb)

    ev.run()

    dt = np.diff(buffer)
    logger.debug(f"{dt.mean()=}, {dt.std()=}, {np.quantile(dt, [0.01, 0.99])=}")
    accuracy_decimals = 3
    precision = 0.001

    # Windows Python 3.10 has poor timer resolution (~15.6ms minimum)
    # Skip precise timing assertions for this configuration
    if IS_WINDOWS_PY310:
        accuracy_decimals = 2
        precision = 0.15
        if dt_s < 0.01:
            logger.warning(
                f"Skipping precise timing assertion for Windows Python 3.10 with dt_s={dt_s}. "
                f"Got mean={dt.mean():.6f}s instead of {dt_s}s due to poor timer resolution."
            )
            pytest.skip(
                "Windows Python 3.10 has insufficient timer resolution for sub-10ms intervals"
            )

    assert (
        np.round(dt.mean() - dt_s, decimals=accuracy_decimals) < precision
    )  # one interval is around 10ms for Windows Python 3.10, so we allow a large margin here
    assert abs(dt.mean() + 3 * dt.std() - dt_s) < dt_s


# TODO: add a test for the delayed callbacks



import logging
import struct

import ujson


# By default logging will send the message as pickle - overwrite the method to send json
class UJsonSocketHandler(logging.handlers.SocketHandler):
    """
    A custom logging handler that sends log records as JSON over a socket.

    This handler extends the `SocketHandler` class to serialize log records using the `ujson` library
    instead of the default `pickle` serialization. The log records are converted to JSON format and sent
    over a socket connection. This handler is useful for logging systems that require JSON-formatted log
    messages.

    Additionally, the `makeSocket` method is overridden to set a custom (short) timeout for socket connections.
    This way, if the log server is not available, the application will not hang for too long trying to connect.

    Methods
    -------
    makePickle(record: logging.LogRecord) -> bytes
        Serialize the log record to a JSON-formatted byte string.
    makeSocket(timeout=1) -> socket.socket
        Create a socket with a specified timeout.
    """

    def makeSocket(self, timeout=1):
        sock = super().makeSocket(timeout=timeout)
        return sock

    def makePickle(self, record: logging.LogRecord) -> bytes:
        try:
            s = ujson.dumps({k: v for k, v in record.__dict__.items()}).encode()

        # In case the record contains an error message, try to convert it to str
        except TypeError as e:
            # Log the serialization issue for debugging
            fallback_logger = logging.getLogger(__name__)
            fallback_logger.debug(
                f"Failed to serialize log record with ujson, converting values to strings: {e}"
            )
            s = ujson.dumps(
                {
                    k: v if isinstance(v, int) else str(v)
                    for k, v in record.__dict__.items()
                }
            ).encode()

        slen = struct.pack(">L", len(s))  # is needed by the servers handler
        return slen + s



import ctypes
import logging
import os
from typing import Callable

import numpy as np
import pylsl
import xmltodict

from dareplane_utils.general.ringbuffer import RingBuffer
from dareplane_utils.logging.logger import get_logger

logger = get_logger(__name__)


class StreamWatcherNotConnected(ValueError):
    pass


def get_streams_names() -> list[str]:
    """Get a list of all available lsl stream names.

    Returns
    -------
    streams : list[str]
        names of all available LSL streams

    """

    return [s.name() for s in pylsl.resolve_streams()]


def pylsl_xmlelement_to_dict(inf: pylsl.StreamInfo) -> dict:
    """
    The pylsl XMLElement is hard to investigate -> cast to a dict for
    simplicity
    """
    return xmltodict.parse(inf.as_xml())


def get_channel_names(inf: pylsl.StreamInfo) -> list[str]:
    """
    Extract channel names from the LSL stream info.

    This function parses the XML metadata of the LSL stream info to extract the channel names.
    If the channel information is not available, it returns default channel names in the format "ch_1", "ch_2", etc.

    Parameters
    ----------
    inf : pylsl.StreamInfo
        The LSL stream info object containing the metadata.

    Returns
    -------
    list[str]
        A list of channel names extracted from the stream info.
    """
    d = pylsl_xmlelement_to_dict(inf)

    # By adding to the xml meta data structure of LSL, if we only add one
    # channel, the data will be a dict instead of a list of dicts

    try:
        if isinstance(d["info"]["desc"]["channels"]["channel"], dict):
            return [d["info"]["desc"]["channels"]["channel"]["label"]]
        else:
            return [
                ch_inf["label"] for ch_inf in d["info"]["desc"]["channels"]["channel"]
            ]
    except TypeError as err:
        logger.debug(f"No channel info - continue with default: ch_1, ch_2,... - {err}")
        return [f"ch_{i + 1}" for i in range(inf.channel_count())]


class StreamWatcher:
    def __init__(
        self,
        name: str = "",
        buffer_size_s: float = 2,
        logger: logging.Logger = logger,
    ):
        """
        Parameters
        ----------
        name : str
            Name tag to identify the manager -> could be same as the LSL stream
            it should be watching
        buffer_size_s : float
            the data buffer size in seconds
        """
        self.name = name
        # this concerns the ringbuffer
        self.buffer_size_s = buffer_size_s

        # the maximum number of samples to pull in one go
        self.chunk_buffer_size = 1024 * 32

        self.stream = None
        self.inlet = None
        self.buffer = None

        # Set after connection
        self.samples: list[list[float]] = []
        self.times: list[float] = []
        self.n_new: int = 0
        self.logger = logger

        # used to expose the ring buffer data structure directly
        self.buffer = np.array([])
        self.buffer_t = np.array([])
        self.last_t = 0
        self.curr_i = 0

        # The update method will be overwritting during _init_buffer
        # fit for the date type of the stream
        self.update: Callable = lambda: None

    def connect_to_stream(self, identifier: dict | None = None):
        """
        Either use the self.name or a provided identifier dict to hook up
        with an LSL stream, they should coincide
        """
        if identifier:
            name = identifier["name"]
            self.name = name
        else:
            name = self.name

        self.streams = pylsl.resolve_byprop("name", name)
        if len(self.streams) > 1:
            print(f"Selecting stream by {name=} was ambigous - taking first")

        self.inlet = pylsl.StreamInlet(self.streams[0])

        self.channel_names = get_channel_names(self.inlet.info())

        self._init_buffer()

        # The first update call will return empty, so do it here already
        self.update()

    def _init_buffer(self):
        if self.streams is None or self.inlet is None:
            raise StreamWatcherNotConnected(
                "StreamWatcher seems not connected, did you call"
                " connect_to_stream() on it?"
            )

        # set a default value for irregular streams
        if self.streams[0].nominal_srate() == 0:
            n_samples = 1000
        else:
            n_samples = int(self.streams[0].nominal_srate() * self.buffer_size_s)

        self.ring_buffer = RingBuffer((n_samples, len(self.channel_names)))

        self._adjust_buffer_data_type()

        # link properties -> for convenience of access
        self.buffer = self.ring_buffer.buffer
        self.buffer_t = self.ring_buffer.buffer_t
        self.last_t = self.ring_buffer.last_t
        self.curr_i = self.ring_buffer.curr_i

        self._define_update_method()

    def _adjust_buffer_data_type(self):

        # -- This is kept as caveat -> always compare against the pylsl.cf_* types
        #    using the inlet.info().channel_format()
        # dtype_map = {
        #    ctypes.c_char_p: "object",
        #    ctypes.c_double: np.float64,  # could also be any int type
        #    ctypes.c_byte: np.int8,
        #    ctypes.c_short: np.int16,
        #    ctypes.c_int: np.int32,
        #    ctypes.c_longlong: np.int64,
        # }

        dtype_map = {
            pylsl.cf_string: "object",
            pylsl.cf_double64: np.float64,
            pylsl.cf_int8: np.int8,
            pylsl.cf_int16: np.int16,
            pylsl.cf_int32: np.int32,
            pylsl.cf_int64: np.int64,
        }

        # default to np.float32 << LSL default
        dtype = dtype_map.get(self.inlet.info().channel_format(), np.float32)

        self.chunk_buffer = np.zeros(
            (self.chunk_buffer_size, len(self.channel_names))
        ).astype(dtype)

        self.ring_buffer.buffer = self.ring_buffer.buffer.astype(dtype)

    def _define_update_method(self):

        if self.inlet.value_type == ctypes.c_char_p:
            self.update = self.update_char_p
        else:
            self.update = self.update_numeric

    def update_numeric(self):
        """Look for new data on the LSL inlet and update the buffer"""

        # This logic works as long as the returned samples are not too many
        # samples, times = self.inlet.pull_chunk()
        # Better use the direct assignment

        _, times = self.inlet.pull_chunk(
            max_samples=self.chunk_buffer_size, dest_obj=self.chunk_buffer
        )

        if len(times) > 0:
            samples = self.chunk_buffer[: len(times), :]
            self.ring_buffer.add_samples(samples, times)
            self.overwriting_samples(samples)
            self.n_new += len(times)

    def update_char_p(self):
        """Look for new data on the LSL inlet and update the buffer, processing for strings"""

        samples, times = self.inlet.pull_chunk(
            # max_samples=self.chunk_buffer_size  -- ignore for strings (usually iregular, relatively rare, string markers), as the larger max leads to a much longer pull time. Potentially needs tuning if larger string payloads are sent.
        )

        if len(times) > 0:
            self.ring_buffer.add_samples(samples, times)
            self.overwriting_samples(samples)
            self.n_new += len(times)

    def overwriting_samples(self, samples):
        self.samples = samples

    def unfold_buffer(self):
        return self.ring_buffer.unfold_buffer()

    def unfold_buffer_t(self):
        # Do hstack here as the time buffer will be of dim (n,1) anyways
        return self.ring_buffer.unfold_buffer_t()

    def disconnect(self):
        """Destroying the inlet will disconnect -> see pylsl.pylsl.py"""
        logger.info(f"Disconnecting from LSL stream - LSLInlet:{self.inlet}")
        del self.inlet
        self.inlet = None


if __name__ == "__main__":
    sw = StreamWatcher("mock_EEG_stream")
    sw.connect_to_stream()



import logging
import subprocess
import time
from pathlib import Path
from logging.handlers import SocketHandler

import psutil
import pytest

from dareplane_utils.logging.logger import get_logger


class TerminationError(Exception):
    pass


@pytest.fixture
def reset_logging():
    """Reset logging state to ensure clean socket handler for server tests.

    This fixture is necessary because:
    1. Socket handlers persist across tests in the same process
    2. A failed connection attempt (no server) leaves the handler in a bad state
    3. We need to recreate handlers to test actual server communication
    """
    # Close and remove all socket handlers from root logger
    root_logger = logging.getLogger()
    for handler in root_logger.handlers[:]:
        if isinstance(handler, SocketHandler):
            if hasattr(handler, "sock") and handler.sock:
                try:
                    handler.sock.close()
                except:
                    pass
                handler.sock = None
            handler.close()
            root_logger.removeHandler(handler)

    # Reset the config applied flag so logging can be reconfigured
    import dareplane_utils.logging.logger as logger_module

    logger_module._config_applied = False

    # Clear any connection warnings to test fresh connections
    for handler in root_logger.handlers[:]:
        if isinstance(handler, SocketHandler) and hasattr(
            handler, "_connection_warned"
        ):
            delattr(handler, "_connection_warned")

    yield

    # Cleanup after test
    for handler in root_logger.handlers[:]:
        if isinstance(handler, SocketHandler):
            if hasattr(handler, "sock") and handler.sock:
                try:
                    handler.sock.close()
                except:
                    pass
            handler.close()


def test_opt_out_of_network_logging():
    logger = get_logger("myapp", no_socket_handler=True)
    assert not any([isinstance(h, SocketHandler) for h in logger.handlers])


def run_logging_server() -> subprocess.Popen:
    cmd = "python -m dareplane_utils.logging.server --logfile=dareplane_test.log"
    return subprocess.Popen(cmd, shell=True)


def stop_process_and_children(p: psutil.Process):
    i = 0
    print(f"{p.children()=}")
    for cp in p.children():
        cp.terminate()

    try:
        print(f"{p.pid}-{p.status()=} - after closing children")
        while p.is_running() and not p.status() == "zombie":
            print(f"terminating {p=}")
            p.terminate()
            time.sleep(0.1)
            i += 1
            if i > 10:
                raise TerminationError(f"Cannot stop process {p=} from running")
    except psutil.NoSuchProcess as err:
        print(f"{err=}")
        pass


def test_logging_server(reset_logging):
    testf = Path("dareplane_test.log")

    try:
        with run_logging_server() as p_server:
            print(f"Started logging server with PID {p_server.pid}")
            time.sleep(1)

            print("Initializing loggers and sending messages")
            logger1 = get_logger("myapp.area1")
            print("Initialized first logger")
            logger2 = get_logger("myapp.area2")
            print("Initialized second logger")
            print("Setting log level")
            logger1.setLevel(10)
            logger2.setLevel(10)
            print("Log level set")

            print("Sending messages")
            logger1.debug("debug1")
            logger1.info("info1")
            logger1.warning("warning1")
            logger1.error("error1")
            logger2.debug("debug2")
            logger2.info("info2")
            logger2.warning("warning2")
            logger2.error("error2")
            print("Messages sent")

            time.sleep(1)
            # Check last lines in log file
            print("Reading the logfile")
            with open(testf, "r") as fl:
                ll = fl.readlines()

            # delete the loggers to get rid of file handlers, which might block
            # a proper cleanup
            print("Deleting loggers and sending messages")
            del logger1, logger2

            print("Stopping logging server")
            stop_process_and_children(psutil.Process(p_server.pid))

        assert all(
            [
                e in "".join(ll[:4])
                for e in [
                    "debug1",
                    "info1",
                    "warning1",
                    "error1",
                ]
            ]
        ), f"Log lines do not match expectation for logger1, got {ll[:4]}"

        assert all(
            [
                e in "".join(ll[4:])
                for e in [
                    "debug2",
                    "info2",
                    "warning2",
                    "error2",
                ]
            ]
        ), f"Log lines do not match expectation for logger2, got {ll[4:]}"

    finally:
        try:
            testf.unlink()
        except PermissionError as err:
            print(f"Could not delete test file {testf}, got {err}")
        except FileNotFoundError as err:
            print(f"Log file did not exits - got {err}")


# --------------------- plain performance tests ----------------------
# def mytest():
#     # Assuming there is a logging server running, test performance for writing
#     # to disk vs writing to a socket
#     socketHandler = logging.handlers.SocketHandler(
#         "localhost", logging.handlers.DEFAULT_TCP_LOGGING_PORT
#     )
#     logger1 = logging.getLogger("myapp.socket")
#     logger2 = logging.getLogger("myapp.file")
#
#     logger1.addHandler(socketHandler)
#     logger2.addHandler(fh)
#     logger1.setLevel(10)
#     logger2.setLevel(10)
#
#     logger1.debug("test")
#     logger2.debug("test")
#
#     # %timeit logger1.debug("test")
#     # %timeit logger2.debug("test")
#
#     # In [34]:     %timeit logger1.debug("test")
#     # 39.3 µs ± 2.83 µs per loop (mean ± std. dev. of 7 runs, 10,000
#     # loops each)
#
#     # In [35]:     %timeit logger2.debug("test")
#     # 40.5 µs ± 8.8 µs per loop (mean ± std. dev. of 7 runs, 10,000 l
#     # oops each)



import numpy as np
import pytest
from scipy.signal import sosfilt

from dareplane_utils.signal_processing.filtering import FilterBank


@pytest.fixture
def x():
    sfreq = 1000

    # test with two channels
    x = np.sin(np.linspace(0, 200, 20 * sfreq)).reshape(-1, 1)
    x += np.random.randn(*x.shape) * 0.5
    x = np.hstack([x, x * 2])
    return x


def test_filterbank_vs_manual_filtering(x):

    sfreq = 1000

    fb = FilterBank(
        bands={"alpha": [8, 12], "beta": [12, 30], "gamma": [30, 80]},
        sfreq=sfreq,
        n_in_channels=x.shape[1],
        output="abs_ma",
        filter_buffer_s=2,
        n_lookback=sfreq // 10,
    )

    # single data chunk -> 20k samples, should fill over the buffer
    times = np.arange(x.shape[0]) / sfreq
    fb.filter(x, times)
    xf = fb.ring_buffer.unfold_buffer()

    xf_alpha, _ = sosfilt(fb.sos["alpha"], x, zi=fb.zis["alpha"], axis=0)
    xf_beta, _ = sosfilt(fb.sos["beta"], x, zi=fb.zis["beta"], axis=0)

    assert np.allclose(xf_alpha[-1000:, 0], xf[-1000:, 0, 0])
    assert np.allclose(xf_alpha[-1000:, 1], xf[-1000:, 1, 0])
    assert np.allclose(xf_beta[-1000:, 0], xf[-1000:, 0, 1])
    assert np.allclose(xf_beta[-1000:, 1], xf[-1000:, 1, 1])


def test_filterbank_get_data(x):

    sfreq = 1000
    n_lookback = sfreq // 10

    fb = FilterBank(
        bands={"alpha": [8, 12], "gamma": [30, 80]},
        sfreq=sfreq,
        n_in_channels=x.shape[1],
        output="abs_ma",
        filter_buffer_s=2,
        n_lookback=n_lookback,
    )

    times = np.arange(x.shape[0]) / sfreq
    fb.filter(x, times)

    xf = fb.ring_buffer.unfold_buffer()

    xta = np.mean(np.abs(xf[-n_lookback:, 0, 0]))
    xtag = np.mean(np.abs(xf[-n_lookback:, 0, 1]))
    xtb = np.mean(np.abs(xf[-n_lookback:, 1, 0]))
    xta2 = np.mean(np.abs(xf[-(n_lookback + 10) : -10, 0, 0]))

    # AbsMovingAverage -> the default
    mva = fb.get_data()
    assert np.allclose(mva[-1, 0, 0], xta)
    assert np.allclose(mva[-1, 1, 0], xtb)
    assert np.allclose(mva[-11, 0, 0], xta2)
    assert np.allclose(mva[-1, 0, 1], xtag)

    # Signal == noop
    fb.output = "signal"
    fb.output_transform = fb.output_posprocessing_map[fb.output]
    xs = fb.get_data()
    assert np.allclose(xs, xf[-xs.shape[0] :, ...])

    # Square
    fb.output = "square"
    fb.output_transform = fb.output_posprocessing_map[fb.output]
    xs = fb.get_data()
    assert np.allclose(xs, xf[-xs.shape[0] :, ...] ** 2)


def test_filterbank_single_vs_chunk_processing(x):

    sfreq = 1000
    n_lookback = sfreq // 10

    fb = FilterBank(
        bands={"alpha": [8, 12], "gamma": [30, 80]},
        sfreq=sfreq,
        n_in_channels=x.shape[1],
        output="abs_ma",
        filter_buffer_s=2,
        n_lookback=n_lookback,
    )

    fb2 = FilterBank(
        bands={"alpha": [8, 12], "gamma": [30, 80]},
        sfreq=sfreq,
        n_in_channels=x.shape[1],
        output="abs_ma",
        filter_buffer_s=2,
        n_lookback=n_lookback,
    )

    times = np.arange(x.shape[0]) / sfreq

    inc = 100
    times = np.arange(x.shape[0]) / sfreq

    for iseg in range(0, x.shape[0] // 2, inc):
        data = x[iseg : iseg + inc, :]
        t = times[iseg : iseg + inc]
        fb2.filter(data, t)

    fb.filter(x[: iseg + inc], times)

    assert np.allclose(
        fb.ring_buffer.unfold_buffer(), fb2.ring_buffer.unfold_buffer()
    )



import socket
import sys
import threading
import time

import psutil
import pytest

from dareplane_utils.default_server.server import DefaultServer
from dareplane_utils.logging.logger import get_logger
from tests.resources.shared import get_test_subprocess

logger = get_logger("testlogger")
logger.setLevel("DEBUG")

# Platform-specific timeout for process operations
IS_WINDOWS = sys.platform == "win32"
PROCESS_TIMEOUT = 2.0 if IS_WINDOWS else 1.0

# for checking which process is running at a given port, we can use netstat with e.g.
# `netstat -anv -p tcp | grep 8080`


@pytest.fixture
def get_default_server_with_process_spawning():
    # Init the server
    server = DefaultServer(port=8080)
    stop_event = threading.Event()
    server.init_server(stop_event=stop_event)

    # Add process creation capabilities
    pcommand_map = {"STARTPROCESS": get_test_subprocess}
    server.pcommand_map = pcommand_map

    # Run the servers processing in a separate thread
    stop_event.clear()  # in case it was set before
    server_thread = threading.Thread(target=server.start_listening)
    server_thread.start()

    yield server_thread, stop_event, server

    # Teardown
    stop_event.set()
    server_thread.join()
    server.shutdown()


def test_spawning_processes_from_client(
    get_default_server_with_process_spawning,
):
    server_thread, stop_event, server = get_default_server_with_process_spawning

    time.sleep(0.2 if IS_WINDOWS else 0.1)

    # Send a message to the server to spawn a process
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect(("localhost", 8080))

    logger.debug("Sending STARTPROCESS")
    client.sendall(b"STARTPROCESS;")
    time.sleep(0.2 if IS_WINDOWS else 0.1)

    logger.debug(f"{server.processes=}")

    # the process should be registerd for book keeping
    assert len(server.processes.keys()) == 1

    client.sendall(b"CLOSE;")
    client.close()


def test_stopping_processes(get_default_server_with_process_spawning):
    server_thread, stop_event, server = get_default_server_with_process_spawning

    # Send a message to the server to spawn a process
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect(("localhost", 8080))
    client.sendall(b"STARTPROCESS;")

    # the process should be registerd for book keeping
    time.sleep(0.2 if IS_WINDOWS else 0.1)  # allow for thread to process command and spawn
    subp = list(server.processes.values())[0]
    # print(f"{subp=}")
    # print(f"{psutil.Process(subp).children()=}")

    server.close_processes()

    # Give Windows more time to terminate processes
    time.sleep(PROCESS_TIMEOUT)

    assert len(server.processes.keys()) == 0

    # Validate that the process was killed by checking the children
    try:
        parent_ps = psutil.Process(subp.pid)
        assert parent_ps.children() == []
    except psutil.NoSuchProcess:
        # Process already terminated - this is fine
        pass
    
    client.sendall(b"CLOSE;")
    client.close()



import json
import signal
import subprocess
import sys
import threading
import time
from logging import Logger
from typing import Callable

import psutil

from dareplane_utils.logging.logger import get_logger

logger = get_logger(__name__)

# Check if we're on Windows
IS_WINDOWS = sys.platform == "win32"


def noop(*args, **kwargs) -> int:
    return 0


def parse_msg(
    msg: str, pcommand_map: dict, logger: Logger = logger
) -> tuple[Callable, tuple, dict]:
    """
    Parse a bytes msg and return the relevant function + potential kwargs

    Parameters
    ----------
    msg : str
        the msg which arrived at the socket
    pcommand_map : dict
        primary command map to link a <primary_command_str> to a Callable

    Returns
    -------
    func : Callable
        the function to call
    args : tuple
        args to use for the function call
    kwargs : dict
        kwargs to pass to the function

    """

    logger.debug(f"Splitting: {msg=}")
    split = msg.decode().split("|")
    pcomm = split[0]
    args = split[1:-1]
    try:
        kwargs = json.loads(split[-1]) if len(split) > 1 else {}
    except json.JSONDecodeError as e:
        logger.error(
            f"Could not parse json payload {msg.decode()=}: {e}, ignoring msg!"
        )
        return noop, (), {}  # NOOP with no args or kwargs

    return pcommand_map[pcomm], args, kwargs


# This is the default behavior for interpretation of messages
def interpret_msg(
    binary_msg: str, pcommand_map: dict, logger: Logger = logger, **kwargs
) -> threading.Thread | subprocess.Popen | int:
    """Interpret a message and start a threading, subprocess or return an int

    Parameters
    ----------
    binary_msg : str
        the binary string msg as received by the socket
    pcommand_map : dict
        a map of primary commands to functions
    **kwargs
        additonally passed kwargs to the function

    Returns
    -------
    threading.Thread | subprocess.Popen | int
        depending on the implementation of this function in individual modules
        this function could return a thread, subprocess or just integers.
        the server will keep track of subprocesses and threads and will close
        them during the shutdown
    """
    # Get correct function and parse kwargs from json payload
    func, pargs, pkwargs = parse_msg(
        binary_msg, pcommand_map=pcommand_map, logger=logger
    )

    # Add kwargs which might have been passed to the server     # noqa
    pkwargs.update(**kwargs)

    # Note, if func is a decorated function, we would not have a __name__
    # so we are printing the full func= here.
    logger.debug(f"Interpreting {func=}, {pargs=}, {pkwargs=}")
    ret = func(*pargs, **pkwargs)

    return ret


def stop_thread(thread: threading.Thread, logger: Logger = logger):
    """Stopping a thread, the standard way

    Parameters
    ----------
    thread : threading.Thread

    """
    logger.info("Stopping threads")
    # stop_event.set()
    # logger.info("Stop event was set, joining the threads")
    try:
        thread.join()  # clean-up
    except RuntimeError as e:
        if str(e) != "cannot join thread before it is started":
            raise
        else:
            logger.info("Closing thread before it started")


def stop_process(process: subprocess.Popen, logger: Logger = logger):
    """Close all child processes of a Popen instance"""

    parent_ps = psutil.Process(process.pid)
    logger.info(f"Closing all processes of {parent_ps=}")
    max_iter = 5
    i = 0
    j = 0

    # close the children
    while parent_ps and parent_ps.children() != [] and i <= max_iter:
        if i > 0:
            time.sleep(0.2)
        # try gentle termination first
        if i <= max_iter - 1:
            for ch in parent_ps.children():
                logger.debug(f"Sending terminate signal to child process={ch}")
                if IS_WINDOWS:
                    # On Windows, use terminate() method
                    ch.terminate()
                else:
                    # On Unix, use SIGTERM
                    ch.send_signal(signal.SIGTERM)
        # last iteration, if it is not gone yet, kill it
        else:
            for ch in parent_ps.children():
                logger.debug(f"Sending kill to child process={ch}")
                ch.kill()

        i += 1
        logger.debug(f"Remaining childern={parent_ps.children()}")

    # Now handle the parent process
    while (
        psutil.pid_exists(parent_ps.pid)
        and parent_ps.status() != "zombie"
        and j <= max_iter
    ):
        if IS_WINDOWS:
            # On Windows, use terminate() method
            logger.debug(f"Terminating parent={parent_ps.pid}")
            parent_ps.terminate()
        else:
            # On Unix, use SIGINT
            parent_ps.send_signal(signal.SIGINT)
            logger.debug(f"Sent SIGINT to parent={parent_ps.pid}")

        if j > 0:
            time.sleep(0.2)
        j += 1

    # Final kill if still alive
    try:
        parent_ps.kill()
    except psutil.NoSuchProcess:
        # Process already terminated
        pass



import os
import threading
import time

import numpy as np
import pylsl
import pytest

from dareplane_utils.logging.logger import get_logger
from dareplane_utils.stream_watcher.lsl_stream_watcher import StreamWatcher

# from tests.resources.shared import get_test_thread

logger = get_logger("testlogger")


# for checking which process is running at a given port, we can use netstat with e.g.
# `netstat -anv -p tcp | grep 8080`


def provide_lsl_stream(
    stop_event: threading.Event, srate: float = 100, nsamples: int = 1000
):
    outlet = pylsl.StreamOutlet(
        pylsl.StreamInfo("test", "test", 5, 100, "float32", "test")
    )
    data = np.tile(np.linspace(0, 1, nsamples), (5, 1))
    data = data.T * np.arange(1, 6)  # 5 channels with linear increase
    data = data.astype(np.float32)

    isampl = 0
    nsent = 0
    tstart = time.time_ns()
    while not stop_event.is_set():
        dt = time.time_ns() - tstart
        req_samples = int((dt / 1e9) * srate) - nsent
        if req_samples > 0:
            outlet.push_chunk(data[isampl : isampl + req_samples, :].tolist())
            nsent += req_samples
            isampl = (isampl + req_samples) % data.shape[0]  # wrap around

        time.sleep(1 / srate)


@pytest.fixture
def spawn_lsl_stream():

    stop_event = threading.Event()
    stop_event.clear()
    th = threading.Thread(target=provide_lsl_stream, args=(stop_event,))
    th.start()

    yield stop_event

    # teardown
    stop_event.set()
    th.join()


@pytest.fixture
def get_markers_outlet() -> pylsl.StreamOutlet:
    info = pylsl.StreamInfo(
        "TestMarkers",
        "Markers",
        1,
        nominal_srate=pylsl.IRREGULAR_RATE,
        channel_format="string",
    )

    return pylsl.StreamOutlet(info)


def test_connecting(spawn_lsl_stream):
    sw = StreamWatcher("test")
    sw.connect_to_stream()

    # after connect, the number of channels should be known and the buffers
    # should be initialized, with the correct data type
    assert isinstance(sw.inlet, pylsl.StreamInlet)
    assert sw.channel_names == [f"ch_{i}" for i in range(1, 6)]
    assert sw.buffer.shape == (200, 5)
    assert sw.buffer_t.shape == (200,)
    assert sw.chunk_buffer.shape == (32768, 5)
    assert sw.buffer.dtype == np.float32


def test_ringbuffer_updates(spawn_lsl_stream):
    sw = StreamWatcher("test")
    sw.connect_to_stream()

    sw.update()
    time.sleep(0.6)
    sw.update()

    # data should be linearly increasing
    data = sw.unfold_buffer()[-sw.n_new :]
    diff = np.diff(data, axis=0)

    assert sw.n_new > 50

    # diff should be roughly constant and scaled along channels
    assert np.abs(diff - diff[0, :]).max() < 1e-6
    assert np.abs(diff[:, 1] - 2 * diff[:, 0]).max() < 1e-6
    assert np.abs(diff[:, 2] - 3 * diff[:, 0]).max() < 1e-6
    assert np.abs(diff[:, 3] - 4 * diff[:, 0]).max() < 1e-6
    assert np.abs(diff[:, 4] - 5 * diff[:, 0]).max() < 1e-6


def test_reading_string_marker(get_markers_outlet):
    markers_outlet = get_markers_outlet
    # markers_outlet = get_markers_outlet()

    sw = StreamWatcher("TestMarkers")
    sw.connect_to_stream()
    sw.update()

    time.sleep(0.1)

    for i in range(10):
        markers_outlet.push_sample([f"Marker - {i + 1}"])

    time.sleep(0.1)

    # check that the correct update method is selected
    assert sw.update == sw.update_char_p

    sw.update()

    assert sw.buffer[0] == ["Marker - 1"]
    assert sw.buffer[2] == ["Marker - 3"]
    assert sw.buffer[9] == ["Marker - 10"]


@pytest.mark.parametrize(
    "fmt", ["float32", "double64", "string", "int32", "int16", "int8", "int64"]
)
def test_data_format_derivation(fmt):
    # The valid formats are provided in the keys of pylsl.string2fmt
    # string2fmt = {
    #     "float32": cf_float32,
    #     "double64": cf_double64,
    #     "string": cf_string,
    #     "int32": cf_int32,
    #     "int16": cf_int16,
    #     "int8": cf_int8,
    #     "int64": cf_int64,
    # }
    fmt_map = {
        "float32": np.float32,
        "double64": np.float64,
        "string": "object",
        "int32": np.int32,
        "int16": np.int16,
        "int8": np.int8,
        "int64": np.int64,
    }
    sname = f"TestStream_{fmt}"
    info = pylsl.StreamInfo(
        sname,
        "EEG",
        10,
        nominal_srate=100,
        channel_format=fmt,
    )
    outlet = pylsl.StreamOutlet(info)

    if os.name == "nt" and fmt == "int64":
        pass  # pylsl will raise an error saying that int64 is not implemented for windows -> skip this
    else:
        sw = StreamWatcher(sname)
        sw.connect_to_stream()
        sw.update()

        time.sleep(0.1)

        if fmt == "string":
            data = [[f"mrk_{i}" for i in range(10)]] * 5
        else:
            data = np.arange(100).reshape((-1, 10)).astype(fmt_map[fmt])

        outlet.push_chunk(data)

        time.sleep(0.1)

        # check that the correct update method is selected
        if fmt == "string":
            assert sw.update == sw.update_char_p
        else:
            assert sw.update == sw.update_numeric

        sw.update()

        for i, d in enumerate(data):
            assert np.all(sw.buffer[i] == d)



from functools import partial

import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage import convolve
from scipy.signal import butter, sosfilt, sosfilt_zi, sosfiltfilt

from dareplane_utils.general.ringbuffer import RingBuffer

FILTER_MAP = {
    "butter": butter,
}


class FilterBufferOverflow(Exception):
    pass


class FilterBank:
    """
    A class to apply multiple bandpass filters to multi-channel signal data.

    Attributes
    ----------
    bands : dict
        Dictionary of frequency bands to filter.
    order : int
        Order of the filter.
    type : str (default: 'butter')
        Type of filter to use. Currently only 'butter' is implemented. Using the naming convention of scipy.signal.
    sfreq : int
        Sampling frequency of the input signal.
    output : str
        Type of output transformation to apply. Options are 'abs_ma', 'signal', 'square'.
    n_in_channels : int
        Number of input channels.
    filter_buffer_s : float
        Length of the filter buffer in seconds.
    n_lookback : int
        Number of samples to look back for the moving average 'abs_ma' output transformation.

    Methods
    -------
    filter(data, times)
        Apply the filters to the input data and store the results in the ring buffer.

    get_data()
        Retrieve the filtered data from the ring buffer, applying the selected output transformation.
    """

    def __init__(
        self,
        bands: dict[str, list[float, float]],
        order: int = 8,
        type: str = "butter",
        sfreq: int = 1000,
        output: str = "signal",
        n_in_channels: int = 1,
        filter_buffer_s: float = 1,
        n_lookback: int = 5,
    ):
        self.sfreq = sfreq
        self.output = output
        self.n_in_channels = n_in_channels
        self.ch_names = [f"ch_{i}" for i in range(n_in_channels)]
        self.bands = bands  # keep track of the bands for conveniece and checking from the outside

        # Used if output is 'abs_ma'
        self.n_lookback = n_lookback

        self.sos = {
            k: butter(
                order,
                (v[0], v[1]),
                btype="bandpass",
                output="sos",
                fs=self.sfreq,
            )
            for k, v in bands.items()
        }

        self.zis = {
            # create the initial zi for the filter and expand for the number of channels
            k: np.repeat(
                np.expand_dims(sosfilt_zi(v), axis=1),
                self.n_in_channels,
                axis=1,
            )
            for k, v in self.sos.items()
        }

        self.ring_buffer = RingBuffer(
            (int(sfreq * filter_buffer_s), len(self.ch_names), len(bands))
        )

        self.output_posprocessing_map = {
            "abs_ma": partial(abs_moving_average, n_lookback=n_lookback),
            "signal": noop,
            "square": lambda x: x**2,
        }
        self.output_transform = self.output_posprocessing_map[output]
        self.n_new = 0

    def filter(self, data: np.ndarray, times: np.ndarray):
        """Assume incoming data to be n_samples x n_in_channels"""

        # I tried to have this write to the buffer directly without stacking
        # an array here, but this feels like premature optimization
        #
        # TODO: have a look at how numpy's strides and as_strided could
        # be used to write to mulitple slices with one return!
        # --> this concerns more the buffer
        #
        # Also check if using a ring buffer per channel could be faster
        # --> should only make a difference if filter() is called much more often
        # than pushing to LSL.

        # TODO: consider reshaping the zis so that sosfilt can be used with axis=0
        # plain %timeit suggests that we loose about 50ns per transpose...
        # the allocation total is ~400ns for 100 samples with 2 channels
        fdata = np.zeros((data.shape[0], len(self.ch_names), len(self.sos))).T

        for i, (k, sos) in enumerate(self.sos.items()):
            (fdata[i, :, :], self.zis[k]) = sosfilt(sos, data.T, zi=self.zis[k])

        self.ring_buffer.add_samples(fdata.T, times)
        self.n_new += data.shape[0]

    def get_data(self) -> np.ndarray:

        # get some extra history for the MA
        data = self.ring_buffer.unfold_buffer()[-(self.n_new + self.n_lookback) :, ...]

        out = self.output_transform(data)[-self.n_new :, ...]

        return out


def abs_moving_average(
    data: np.ndarray,
    n_lookback: int = 5,
) -> np.ndarray:

    kernel = np.ones((n_lookback, *([1] * len(data.shape[1:])))) * 1 / n_lookback

    ma = convolve(np.abs(data), kernel, mode="constant", cval=0.0)

    # Cut off the edges, so that we return the same as would be for non-nan
    # pandas.rolling(n_lookback).mean()
    pre = (n_lookback - 1) // 2
    post = n_lookback // 2

    return ma[pre:-post]


def noop(data: np.ndarray, *args) -> np.ndarray:
    return data


if __name__ == "__main__":
    sfreq = 1000
    x = np.sin(np.linspace(0, 200, 20 * sfreq)).reshape(-1, 1)
    x += np.random.randn(*x.shape) * 0.5
    x = np.hstack([x, x * 2])

    fb = FilterBank(
        bands={"alpha": [8, 12], "beta": [12, 30], "gamma": [30, 80]},
        sfreq=sfreq,
        n_in_channels=x.shape[1],
        output="abs_ma",
        # output="signal",
        filter_buffer_s=2,
        n_lookback=sfreq // 10,
    )

    rets = []
    inc = 100
    times = np.arange(x.shape[0]) / sfreq

    for iseg in range(0, x.shape[0] // 2 - 5 * inc, inc):
        data = x[iseg : iseg + inc, :]
        t = times[iseg : iseg + inc]
        fb.filter(data, t)

        if fb.ring_buffer.curr_i > 200:
            rets.append(fb.get_data())

    b = np.cumsum([r.shape[0] for r in rets])
    xf = np.vstack(rets)

    nstart = 300
    nend = len(x)
    fig, axs = plt.subplots(3, 1, sharex=True)
    axs[0].plot(x[nstart:nend, 0])
    axs[1].plot(xf[nstart:nend, 0], "r.-", label=fb.ch_names[0])
    axs[1].plot(b - 1 - nstart, xf[b - 1, 0], "b.")
    axs[2].plot(xf[nstart:nend, 1], "r.-", label=fb.ch_names[1])

    # add filtfilt results
    xff = np.hstack([sosfiltfilt(sos, x, axis=0) for sos in fb.sos.values()])
    axs[1].plot(xff[nstart:nend, 0], "g.-", label=fb.ch_names[0] + "_filtfilt")
    axs[2].plot(xff[nstart:nend, 1], "g.-", label=fb.ch_names[1] + "_filtfilt")

    # add filt but whole chunk at once
    zis = {
        k: np.repeat(
            np.expand_dims(sosfilt_zi(v), axis=1),
            1,
            axis=1,
        )
        for k, v in fb.sos.items()
    }
    xfonce = np.hstack(
        [sosfilt(fb.sos[bn], x.T, axis=1, zi=zi)[0].T for bn, zi in zis.items()]
    )

    axs[1].plot(xfonce[nstart:nend, 0], "b.-", label=fb.ch_names[0] + "_filt")
    axs[2].plot(xfonce[nstart:nend, 1], "b.-", label=fb.ch_names[1] + "_filt")

    axs[1].legend()
    axs[2].legend()

    plt.show()

    # import pandas as pd
    #
    # def foo(data, n_lookback):
    #     df = pd.DataFrame(np.abs(data))
    #     ma = df.rolling(n_lookback).mean()
    #     return ma.loc[~ma[0].isna(), :]
    #
    # data = np.arange(30).reshape(10, 3)
    #
    # %timeit foo(data, 5)
    #
    # %timeit abs_moving_average(data, 5)

    # In [70]:     %timeit foo(data, 5)
    # 156 µs ± 217 ns per loop (mean ± std. dev. of 7 runs, 10,000 loops each)
    #
    # In [71]:     %timeit abs_moving_average(data, 5)
    # 7.42 µs ± 71.5 ns per loop (mean ± std. dev. of 7 runs, 100,000 loops each)



import time


def sleep_s(s: float, partial_sleep_threshold: float = 0.0005, nsteps: int = 30):
    """
    Sleep for a specified duration with partial sleep optimization.

    Parameters
    ----------
    s : float
        The total duration to sleep in seconds.
    partial_sleep_threshold : float, optional
        The threshold duration above which partial sleep optimization is applied, by default 0.0005.
        I.e., only for durations `s` above the threshold, the optimization is applied.
    nsteps : int, optional
        The number of steps for partial sleep, by default 30. Empirical testing showed very good accuracy for 30.
        If you want to optimize for CPU load, reduce to `nsteps` > 4.
    """

    start = time.perf_counter()
    if s >= partial_sleep_threshold:
        partial_sleep(s, start, nsteps)

    # Sleep for the remaining time
    full_speed(s, start)


def partial_sleep(s: float, start: float, nsteps: int = 30):
    """Sleep for 90% of `s` or up to 30ms to the end, whatever is longer"""

    assert nsteps > 4, (
        "Empirical tests showed that irrespective of the size of `s`, sleeping"
        " seemes to maximally overshoot by a factor of 4. Keep nsteps larger "
        "than 4 to be on the safe side."
    )

    s_max = max(s - 0.03, s * 0.9)

    while time.perf_counter() - start < s_max:
        time.sleep(s / nsteps)


def full_speed(s: float, start: float):
    while time.perf_counter() - start < s:
        pass



# A plain event_loop which precisely adheres to timing by using the adjusted sleep_s
import inspect
import time
from dataclasses import dataclass
from threading import Event
from typing import Any, Callable, Optional

from dareplane_utils.general.time import sleep_s


@dataclass
class DelayedCallback:
    """
    Simple auxiliary struct to store a callback and a delay.

    This class is used to store a callback function along with a delay and a start timestamp.
    It is typically used in the context of an event loop to schedule callbacks to be executed
    after a specified delay.

    Attributes
    ----------
    cb : Callable
        The callback function to be executed.
    delay_s : float
        The delay in seconds before the callback is executed relative to `self.start_ts`.
    start_ts : float
        The timestamp when the delay starts. Defaults to the current time.
    """

    cb: Callable
    delay_s: float
    start_ts: float = time.perf_counter()


class EventLoop:
    """
    A class that implements a custom event loop with precise timing.

    The EventLoop uses dareplane_utils.general.time.sleep_s for more precise
    sleep timing at the expense of CPU usage.

    Callbacks are the means of interacting with the event loop. There are two types of callbacks:

    - Periodic callbacks: These are executed at regular intervals.

    - One-time callbacks: These are executed once and then removed from the list of callbacks.
      One-time callback can furthermore be scheduled to run at a specific time in the future.

    Callbacks can be any callable function, which usually gets one argument `ctx` -
    a context object, that can be of type any. This ensures that any type of input can
    be implemented. See the example section for more details.


    Attributes
    ----------
    dt_s : float
        The time interval in seconds between callback executions.
    stop_event : threading.Event
        An event that, when set, signals the event loop to stop.
    last_run_ts : float
        The timestamp of the last callback execution.
    callbacks : list[Callable]
        A list of periodic callback functions.
    callbacks_once : list[Callable]
        A list of one-time callback functions.
    delayed_callbacks_once : list[DelayedCallback]
        A list of delayed one-time callback functions.
    ctx : Any
        A context object that is passed to the callback functions.
    now : float
        The current timestamp as of time.perf_counter(). Used for time keeping internally.


    Examples
    --------
    >>> def no_arg_callback():
    ...     print("Running with no args")
    >>>
    >>> evloop = EventLoop(dt_s=0.1)  # process callbacks every 100ms
    >>> evloop.add_callback_once(lambda ctx: no_arg_callback())
    >>>
    >>> # just to stop the loop after 1s
    >>> evloop.add_delayed_callback_once(cb=lambda ctx: evloop.stop_event.set(), dt=1)
    >>>
    >>> evloop.run()
    Running with no args
    >>>
    >>> # NOTE: This example is left to explain how wrapping works. Creating a wrapper is no longer necessary
    >>> #       if no `ctx` is an arg / kwargs, the event_loop.validate_callback() will automatically create
    >>> #       a wrapper version for you.

    >>> def custom_arg_and_kwarg_callback(a, b=1, c=2):
    ...     print(f"Running with {a=}, {b=}, {c=}")
    >>>
    >>> def wrapped_custom_arg_and_kwargs(ctx: dict):
    ...     custom_arg_and_kwarg_callback(ctx["a"], b=ctx["b"], c=ctx["c"])
    >>>
    >>> evloop = EventLoop(dt_s=0.1, ctx={"a": 11, "b": 22, "c": 33})
    >>> evloop.add_callback_once(wrapped_custom_arg_and_kwargs)
    >>>
    >>> # just to stop the loop after 1s
    >>> evloop.add_delayed_callback_once(cb=lambda ctx: evloop.stop_event.set(), dt=1)
    >>>
    >>> evloop.run()
    Running with a=11, b=22, c=33

    """

    def __init__(
        self, dt_s: float, stop_event: Optional[Event] = None, ctx: Optional[Any] = None
    ):
        """
        Initialize the EventLoop with a specified time interval and optional context.

        Parameters
        ----------
        dt_s : float
            The time interval in seconds between callback executions.
        stop_event : Optional[Event], optional
            An event that, when set, signals the event loop to stop. If not provided, a new Event is created.
        ctx : Optional[Any], optional
            A context object that is passed to the callback functions. Defaults to None. The self.ctx will be
            passed to every callback regardless of value
        """

        self.dt_s = dt_s
        self.stop_event = stop_event if stop_event is not None else Event()
        self.last_run_ts = time.perf_counter()
        self.callbacks = []
        self.callbacks_once = []
        self.delayed_callbacks_once = []
        self.ctx: Any = ctx
        self.now: float = 0.0

    def add_callback(self, cb: Callable):
        """
        Add a periodic callback to the event loop.

        Parameters
        ----------
        cb : Callable
            The callback function to be added.
        """
        cb = self.validate_callback(cb)

        self.callbacks.append(cb)

    def add_callbacks(self, cbs: list[Callable]):
        """
        A convenience function to add multiple periodic callbacks to the event loop.

        Parameters
        ----------
        cbs : list[Callable]
            A list of callback functions to be added.
        """
        for cb in cbs:
            self.add_callback(cb)

    def add_delayed_callback_once(self, cb: Callable, dt: float = 0.0):
        """
        Add a one-time callback to the event loop that is evaluated after a delay.

        Parameters
        ----------
        cb : Callable
            The callback function to be added.
        dt : float, optional
            The delay in seconds before the callback is executed. Defaults to 0.0.
        """

        cb = self.validate_callback(cb)
        dcb = DelayedCallback(cb, dt, time.perf_counter())
        self.delayed_callbacks_once.append(dcb)

    def add_callback_once(self, cb: Callable):
        """
        Add a one-time callback to the event loop.

        Parameters
        ----------
        cb : Callable
            The callback function to be added.
        """

        cb = self.validate_callback(cb)
        self.callbacks_once.append(cb)

    def add_callbacks_once(self, cbs: list[Callable]):
        """
        Convenience function to add multiple one-time callbacks to the event loop.

        Parameters
        ----------
        cbs : list[Callable]
            A list of callback functions to be added.
        """

        for cb in cbs:
            self.add_callback_once(cb)

    def validate_callback(self, cb) -> Callable:
        """
        Check that every callback accepts at least a kwarg with 'ctx'.
        More kwargs are possible.
        """
        assert isinstance(cb, Callable), f"The provided {cb=} is not a Callable"
        spec = inspect.getfullargspec(cb)

        if "ctx" not in spec.args:  # make a context optional
            return lambda ctx: cb()

        return cb

    def run(self):
        """
        Run the event loop, evaluating the callback every `self.dt_s` seconds
        """
        while not self.stop_event.is_set():

            self.now = time.perf_counter()
            dt_last = self.now - self.last_run_ts > self.dt_s
            if dt_last:
                self.process_callbacks()

                # check how much still needed to sleep as the callback processing
                # might have taken significant amount of time
                dt_remaining = max(0, self.dt_s - (self.now - self.last_run_ts))
                self.last_run_ts = self.now

                sleep_s(dt_remaining)
            else:
                sleep_s(self.dt_s - dt_last)

    def process_callbacks(self):

        while self.callbacks_once != []:
            cb = self.callbacks_once.pop(0)
            cb(ctx=self.ctx)

        if self.delayed_callbacks_once != []:
            to_pop = []
            for i, dcb in enumerate(self.delayed_callbacks_once):
                if self.now - dcb.start_ts > dcb.delay_s:
                    dcb.cb(ctx=self.ctx)
                    to_pop.append(i)

            self.delayed_callbacks_once = [
                dcb
                for i, dcb in enumerate(self.delayed_callbacks_once)
                if i not in to_pop
            ]

        for cb in self.callbacks:
            cb(ctx=self.ctx)


if __name__ == "__main__":
    # examples for use of callbacks

    def no_arg_callback():
        print("Running with no args")

    evloop = EventLoop(dt_s=0.1)  # process callbacks every 100ms

    # for a callback with no args we use lambda to blank the callback arg
    evloop.add_callback_once(lambda ctx: no_arg_callback())

    # stop the evloop after 1s
    evloop.add_delayed_callback_once(cb=lambda ctx: evloop.stop_event.set(), dt=1)

    evloop.run()

    def custom_arg_and_kwarg_callback(a, b=1, c=2):
        print(f"Running with {a=}, {b=}, {c=}")

    # reset to showcase the args
    def wrapped_custom_arg_and_kwargs(ctx: dict):
        custom_arg_and_kwarg_callback(ctx["a"], b=ctx["b"], c=ctx["c"])

    evloop = EventLoop(dt_s=0.1, ctx={"a": 11, "b": 22, "c": 33})
    evloop.add_callback_once(wrapped_custom_arg_and_kwargs)
    # stop the evloop after 1s
    evloop.add_delayed_callback_once(cb=lambda ctx: evloop.stop_event.set(), dt=1)
    evloop.run()
